/*
 * Notification support.
 *
 * Copyright (C) 2021, 2022 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <ipc/cap_alloc.h>

#include "notification.h"
#include "notifier_client.h"

#include "resource_server.h"



/* Initialise endpoints and flags for notifications. */

NotificationSupport::NotificationSupport(unsigned int endpoints)
: _min_endpoints(endpoints)
{
    _endpoints.resize(_min_endpoints);
}

NotificationSupport::~NotificationSupport()
{
}

/* Subscribe to notifications using a notification object, returning the
   endpoint number. */

long NotificationSupport::subscribe(l4_cap_idx_t notifier, notify_flags_t flags,
                                    void *subscriber, unsigned int *endpoint_number)
{
    std::lock_guard<std::mutex> guard(_lock);

    unsigned int endpoint = _endpoints.size();
    _endpoints.resize(endpoint + 1);

    /* Propagate deferred flags for new endpoints. */

    if (_deferred)
        _endpoints[endpoint].deferred = _deferred;

    *endpoint_number = endpoint;
    return _subscribe(endpoint, notifier, flags, subscriber);
}

/* Subscribe to a specific endpoint's notifications using a notification
   object. */

long NotificationSupport::subscribe(unsigned int endpoint, l4_cap_idx_t notifier,
                                    notify_flags_t flags, void *subscriber)
{
    std::lock_guard<std::mutex> guard(_lock);

    if (endpoint >= _endpoints.size())
    {
        _endpoints.resize(endpoint + 1);

        /* Propagate deferred flags for new endpoints. */

        if (_deferred)
            _endpoints[endpoint].deferred = _deferred;
    }

    return _subscribe(endpoint, notifier, flags, subscriber);
}

/* Unsubscribe from an endpoint's notifications. */

void NotificationSupport::unsubscribe(unsigned int endpoint, void *subscriber)
{
    std::lock_guard<std::mutex> guard(_lock);

    _unsubscribe(endpoint, subscriber);
}

/* Common subscription functionality. */

long NotificationSupport::_subscribe(unsigned int endpoint, l4_cap_idx_t notifier,
                                     notify_flags_t flags, void *subscriber)
{
    /* Record details of the notifier itself. */

    NotificationEndpoint &ep = _endpoints[endpoint];

    ep.notifiers.insert(notifier);
    ep.flags = flags;

    /* Send deferred conditions held from before subscription occurred. */

    if (ep.deferred)
    {
        _notify(endpoint, ep.deferred);
        ep.deferred = 0;
    }

    /* Record the subscriber's notifier. */

    _subscribers[subscriber] = notifier;
    return L4_EOK;
}

void NotificationSupport::_unsubscribe(unsigned int endpoint, void *subscriber)
{
    if (endpoint >= _endpoints.size())
        return;

    /* Find and remove the subscriber's notifier. */

    SubscriberNotifierMap::iterator itp = _subscribers.find(subscriber);
    l4_cap_idx_t notifier;

    if (itp == _subscribers.end())
        return;

    notifier = itp->second;
    _subscribers.erase(itp);

    /* Release the notifier. */

    NotificationEndpoint &ep = _endpoints[endpoint];
    NotifierSet::iterator it = ep.notifiers.find(notifier);

    if (it != ep.notifiers.end())
    {
        ep.notifiers.erase(it);
        ipc_cap_free_um(notifier);

        if (ep.notifiers.empty())
        {
            ep.flags = 0;
            ep.deferred = 0;
        }
    }
}

/* Notify a particular endpoint. */

void NotificationSupport::notify(unsigned int endpoint, notify_flags_t flags)
{
    std::lock_guard<std::mutex> guard(_lock);

    _notify(endpoint, flags);
}

void NotificationSupport::_notify(unsigned int endpoint, notify_flags_t flags)
{
    if (endpoint >= _endpoints.size())
        return;

    NotificationEndpoint &ep = _endpoints[endpoint];

    /* Notify the endpoint or hold any notification for potential future
       subscription. */

    if (!ep.notifiers.empty())
    {
        if (flags & ep.flags)
        {
            NotifierSet::iterator it;

            for (it = ep.notifiers.begin(); it != ep.notifiers.end(); it++)
            {
                client_Notifier notifier(*it);

                notifier.notify(flags & ep.flags);
            }
        }
    }
    else
        ep.deferred = flags;
}

/* Notify all endpoints. */

void NotificationSupport::notify_all(notify_flags_t flags)
{
    std::lock_guard<std::mutex> guard(_lock);

    for (unsigned int i = 0; i < _endpoints.size(); i++)
        _notify(i, flags);

    _deferred |= flags;
}

/* Notify the other endpoints. */

void NotificationSupport::notify_others(unsigned int endpoint, notify_flags_t flags)
{
    std::lock_guard<std::mutex> guard(_lock);

    for (unsigned int i = 0; i < _endpoints.size(); i++)
        if (i != endpoint)
            _notify(i, flags);

    _deferred |= flags;
}

/* Release notifiers for each endpoint. */

void NotificationSupport::release_notifiers()
{
    std::lock_guard<std::mutex> guard(_lock);

    for (unsigned int endpoint = 0; endpoint < _endpoints.size(); endpoint++)
    {
        NotificationEndpoint &ep = _endpoints[endpoint];
        NotifierSet::iterator it;

        for (it = ep.notifiers.begin(); it != ep.notifiers.end(); it++)
            ipc_cap_free_um(*it);

        ep.notifiers.clear();
        ep.flags = 0;
        ep.deferred = 0;
    }
}

// vim: tabstop=4 expandtab shiftwidth=4
